"""
Contains the AnnotationExporter class which contains the logic for exporting annotations
"""
import os
import logging as lg
from sqlite3 import connect, Connection, Cursor
import PyPDF2
import PyPDF2.generic
import openpyxl as pyxl
from openpyxl.worksheet.worksheet import Worksheet
from openpyxl.styles import PatternFill
from openpyxl.cell.cell import Cell
from .generic import PDF, Annotation, Page

class AnnotationExporter:
    """
    Responsible for exporting annotations from a pdf to an \n 
    excel file
    """
    def __init__(self) -> None:
        """
        initializes variables for use in the programm
        """
        self.green_cell_fill = PatternFill(
            start_color = "FF00FF00",
            end_color = "FF00FF00",
            fill_type = "solid")
        self.reset_cell_fill = PatternFill(
            fill_type=None,
            start_color="FFFFFFFF",
            end_color="FF000000")
        self.wb: pyxl.Workbook
        self.pdf: PDF
        self.output_folder: str
        self.exporter_col_ds: str | None
        self.exporter_col_var: str | None
        self.ws_datasets: Worksheet
        self.ws_variables: Worksheet
        self.supp_var_names: list[str] = ["QVAL", "QNAM", "QLABEL"]
        self.ds_replace_annots: list[dict] = []
        self.current_page: Page
        lg.basicConfig(
            filename=f"{os.path.dirname(__file__)}/Annotation_Exporter.log",
            encoding="utf-8",
            level=lg.DEBUG,
            filemode="w")

    def determine_exporter_col(self, sheet: str) -> str | None:
        """
        Determines the exporter column. Returns the exel column index of the free column or None.
        The column chosen is the first free column in the sheet. The exporter column is used for 
        marking the entries as present.

        :param sheet: name of the sheet
        :type sheet: str
        :return: the exel column index of the free column or None
        :rtype: str | None
        """
        value = None
        for cell in self.wb[sheet]["1"]:
            if cell.value is None:
                cell.value = "Present in aCRF"
                value =  "".join([i for i in cell.coordinate if not i.isdigit()])  # remove all digits

        if value is None:
            lg.critical("no free column for sheet %s found, exiting...", sheet)
            exit()

        return value

    def export_annotations(self, template_path: str, pdf_path: str, output_folder: str) -> None:
        """
        Exports annots, this is the main function that should be called. 
        Expects the paths to have the correct endings (.pdf, .xlsx).

        :param template_path: path to the template file
        :type template_path: str
        :param pdf_path: path to the pdf file
        :type pdf_path: str
        :param output_folder: path to the output folder
        :type output_folder: str
        """
        print("exporting annotations...")
        lg.info("export annots")
        self.wb = pyxl.load_workbook(template_path)
        self.output_folder = output_folder

        self.exporter_col_ds = self.determine_exporter_col("Datasets")
        self.exporter_col_var = self.determine_exporter_col("Variables")

        self.ws_datasets = self.wb["Datasets"]
        self.ws_variables = self.wb["Variables"]

        self.pdf: PDF = PDF(PyPDF2.PdfReader(pdf_path))

        for page in self.pdf.pages:
            self.current_page = page
            lg.info("starting on page: %s", page.get_page_nr())

            self.add_to_workbook(page.get_annotations())

            lg.debug(page.get_datasets())
            lg.info("Page %s done!", page.get_page_nr())


        self.wb.save(f"{output_folder}/output.xlsx")
        print("generating csv...")
        lg.info("generating csv of export")
        self.generate_variable_csv()
        self.generate_dataset_csv()

        print("complete!")
        lg.info("exported annots")

    def generate_sqlite(self, output_folder: str) -> None:
        """
        generates an sqlite database from the annotations and
        saves it in the output folder

        :param output_folder: path to the output folder
        :type output_folder: str
        """
        conn: Connection= connect(f"{output_folder}/annotations.sqlite")
        c: Cursor = conn.cursor()
        c.execute("DROP TABLE IF EXISTS annotations")
        c.execute("""CREATE TABLE annotations
                    (id INTEGER PRIMARY KEY AUTOINCREMENT,
                    dataset BOOLEAN,
                    new_dataset BOOLEAN,
                    dataset_name TEXT,
                    supp BOOLEAN,
                    assigned_dataset TEXT,
                    variable_name TEXT,
                    content TEXT,
                    color TEXT,
                    page_number INTEGER)""")

        for page in self.pdf.pages:
            for annot in page.get_annotations():
                if annot.is_valid:
                    c.execute("""INSERT INTO annotations
                        (dataset, new_dataset, dataset_name, supp, assigned_dataset, variable_name, content, color, page_number)
                        VALUES (?,?,?,?,?,?,?,?,?)""",
                        (annot.dataset,
                         annot.new_datset,
                         annot.dataset_name,
                         annot.supp,
                         annot.assigned_dataset,
                         annot.variable_name,
                         annot.content,
                         str(annot.color),
                         annot.page.get_page_nr() + 1))

        conn.commit()

    def enter_dataset(self, annot: Annotation) -> None:
        """
        Adds a dataset to the Datasets shhet in the workbook

        :param annot: annotation object
        :type annot: Annotation
        """
        for cell in self.ws_datasets.iter_rows(max_col=1):
            cell = cell[0]
            if cell.value == annot.dataset_name:
                y_coordinate = cell.coordinate.split("A", 1)[1]
                self.ws_datasets[f"{self.exporter_col_ds}{y_coordinate}"] = "Present"
                self.ws_datasets[f"{self.exporter_col_ds}{y_coordinate}"].fill = self.green_cell_fill

                lg.debug(
                    "%s was assigned as a dataset with the color %s",
                    annot.dataset_name, annot.color)
                return

        self.ws_datasets.append({
            "A": annot.dataset_name,
            self.exporter_col_ds: "Present",
        })
        self.ws_datasets[self.exporter_col_ds][self.ws_datasets.max_row - 1].fill = self.green_cell_fill
        self.ws_datasets["A"][self.ws_datasets.max_row - 1].fill = self.reset_cell_fill

    def enter_supp(self, annot: Annotation) -> None:
        """
        adds a SUPPxx dataset to datasets and the three SUPPxx variables to variables

        :param annot: annotation object
        :type annot: Annotation
        """
        self.enter_dataset(annot)

        modified_cols: list[str] = ["B", "C", "L", "F"]

        all_values = [x[0] for x in self.ws_variables.iter_rows(min_col=2, max_col=2, values_only=True)] # very performance inefficient

        if annot.dataset_name in all_values:
            for cell in self.ws_variables["B"]:
                if cell.value != annot.dataset_name:
                    continue
                self.add_page_cell(self.ws_variables[f"M{cell.row}"])
                break
        else:
            for var_name in self.supp_var_names:
                self.ws_variables.append({
                    "B": annot.dataset_name,
                    "C": var_name,
                    "L": "CRF",
                    "F": "200",
                    self.exporter_col_var: "Present"})

                self.add_page_cell(self.ws_variables[f"M{self.ws_variables.max_row}"])
                self.ws_variables[self.exporter_col_var][self.ws_variables.max_row - 1].fill = self.green_cell_fill

                for col in modified_cols:
                    self.ws_variables[col][self.ws_variables.max_row - 1].fill = self.reset_cell_fill

    def add_to_workbook(self, annotations: list[Annotation]) -> None:
        """
        adds both datasets and variables to the workbook.
        Takes a list of annotations, desigend to work with PDF.pages
        
        :param data: list of annotations
        :type data: list[Annotation]
        """
        for annot in annotations:
            if annot.dataset:
                self.enter_dataset(annot)
            elif annot.supp:
                self.enter_supp(annot)
            else:
                self.enter_variable(annot)

    def enter_variable(self, annot: Annotation) -> None:
        """
        adds a variable to the workbook

        :param annot: annotation object
        :type annot: Annotation
        """
        annot.sort_into_datasets()
        if annot.assigned_dataset is None:
            return

        for cell in self.ws_variables["B"]:
            y_coordinate = cell.coordinate.split("B", 1)[1]

            if cell.value == annot.assigned_dataset and self.ws_variables["C" + y_coordinate].value == annot.variable_name:
                if self.ws_variables[f"{self.exporter_col_var}{y_coordinate}"].value == "Present":
                    self.add_page_cell(self.ws_variables[f"M{y_coordinate}"])
                    return

                self.ws_variables[f"{self.exporter_col_var}{y_coordinate}"].value = "Present"
                self.ws_variables[f"{self.exporter_col_var}{y_coordinate}"].fill = self.green_cell_fill
                self.ws_variables[f"L{y_coordinate}"].value = "CRF"
                self.add_page_cell(self.ws_variables[f"M{y_coordinate}"])
                return

        self.ws_variables.append({
            "B": annot.assigned_dataset,
            "C": annot.variable_name,
            "L": "CRF",
            "F": "200",
            self.exporter_col_var: "Present"
        })
        self.add_page_cell(self.ws_variables[f"M{self.ws_variables.max_row}"])
        self.ws_variables[self.exporter_col_var][self.ws_variables.max_row - 1].fill = self.green_cell_fill
        for modified_col in ["B", "C", "L", "F"]:
            self.ws_variables[modified_col][self.ws_variables.max_row - 1].fill = self.reset_cell_fill

    def generate_variable_csv(self) -> None:
        """
        generates the csv for the variables and saves it in the output folder
        """
        csv_list = ["Variable Name#Variable Label#Dataset Name#Page(s)\n"] # start with first line
        for cell in self.ws_variables[self.exporter_col_var]:
            if cell.value == "Present":
                cell_str: str = ""
                cell_str += str(self.ws_variables["C" + str(cell.row)].value) + "#"
                cell_str += str(self.ws_variables["D" + str(cell.row)].value) + "#"
                cell_str += str(self.ws_variables["B" + str(cell.row)].value) + "#"
                cell_str += str(self.ws_variables["M" + str(cell.row)].value) + "\n"
                csv_list.append(cell_str)


        csv_str = "".join(csv_list)
        with open(f"{self.output_folder}/Variables.csv", "w", encoding="utf-8") as f:
            f.write(csv_str)

    def generate_dataset_csv(self):
        """
        generates the csv for the datasets and saves it in the output folder
        """
        csv_list: list[str] = ["Dataset Name#Color\n"] # start with first line
        for page in self.pdf.pages:
            for dataset in page.get_datasets():
                csv_entry: str = f"{dataset[0]}#{dataset[1]}\n"
                if csv_entry in csv_list:
                    continue
                csv_list.append(csv_entry)

        csv_str = "".join(csv_list)

        with open(f"{self.output_folder}/Datasets.csv", "w", encoding="utf-8") as f:
            f.write(csv_str)

    def add_page_cell(self, cell: Cell):
        """
        Appends current page number to the string in a cell if it is not already present.

        :param cell: The cell in which to add the current page number
        :type cell: Cell
        """
        if not cell.value:
            new_value = str(self.current_page.get_page_nr() + 1)
            cell.value = new_value
            cell.fill = self.reset_cell_fill
        elif str(self.current_page.get_page_nr() + 1) in cell.value:
            return
        else:
            new_value = f"{cell.value} {self.current_page.get_page_nr() + 1}"



"""
This is the main file for the annotation export project. It allows the user to export annotations
from a pdf to an excel file. For further information on how to use this please consult the 
README.md
"""
import FreeSimpleGUI as sg
from .annot_export import AnnotationExporter


def ending_present(string: str, ending: str) -> bool:
    """
    checks the presence of the correct ending

    :param string: the string to check
    :type string: str
    :param ending: the ending to check for
    :type ending: str
    :return: True if the ending is present, False otherwise
    :rtype: bool
    """
    split_str = string.split(".")
    return split_str[-1] == ending

def run() -> None:
    """
    runs the gui from the presentation
    """
    annotation_exporter: AnnotationExporter = AnnotationExporter()
    sg.theme("DarkGrey5")
    layout: list[list[sg.Element]] = [
        [sg.Text("Annotation Exporter V 0.2")],
        [sg.Push(), sg.Text("PDF Path"), sg.InputText(key="pdf",default_text=r"PDF/example_compressed"), sg.FileBrowse(file_types=(("PDF", "*.pdf"),))],
        [sg.Push(), sg.Text("Spreadsheet Path"), sg.InputText(key="xlsx", default_text=r"Templates/temp"), sg.FileBrowse(file_types=(("Excel", "*.xlsx"),))],
        [sg.Push(), sg.Text("Output Folder"), sg.InputText(key="output", default_text=r"outputs"), sg.FolderBrowse()],
        [sg.Checkbox("Convert Old Standard", key="conv_old"), sg.Checkbox("Create SQLite database", key="sqlite"), sg.Push(), sg.Button("Export Annotations", key="export")]
        ]

    window: sg.Window = sg.Window(title="Annotation Export", layout=layout, margins=(150, 125))



    while True:
        event, values = window.read()
        if event == sg.WIN_CLOSED:
            break
        try:
            pdf_p = values["pdf"]
            xlsx_p = values["xlsx"]
            output_p = values["output"]
            convert_old = values["conv_old"]
            sqlite = values["sqlite"]
        except KeyError:
            continue

        conv_paths(pdf_p)
        conv_paths(xlsx_p)
        conv_paths(output_p)

        if event == "export":
            if ending_present(xlsx_p, "xlsx"):
                xlsx_path = xlsx_p
            else:
                xlsx_path = f"{xlsx_p}.xlsx"
            if ending_present(pdf_p, "pdf"):
                pdf_path = pdf_p
            else:
                pdf_path = f"{pdf_p}.pdf"

            output_folder = output_p
        else:
            continue

        annotation_exporter.export_annotations(
            xlsx_path,
            pdf_path,
            output_folder)

        if convert_old:
            annotation_exporter.pdf.convert_old_standard(output_folder)

        if sqlite:
            annotation_exporter.generate_sqlite(output_folder)


    window.close()

def conv_paths(path: str) -> None:
    """
    replaces all backslashes to prevent accidental escape sequences

    :param path: the path to replace
    :type path: str
    """
    backslash = r"\ "
    backslash = backslash.split(" ", maxsplit=1)[0] # remove the space after the backslash
    path.replace(backslash, "/")

if __name__ == "__main__":
    run()



from setuptools import setup, find_packages

VERSION = "1.0.7"
DESCRIPTION = "Export eCRF annotations to Excel"
LONG_DESCRIPTION = "Automates migrating between data levels (in this case CRF and STDM spec)"

setup(
    name="annotation_exporter",
    version=VERSION,
    author="Salomo Sparna",
    author_email="SalomoSparna@gmail.com",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    install_requires=["openpyxl", "PyPDF2", "FreeSimpleGUI"],
    keywords=["python", "CRF", "CDISC"],
    classifiers= [
            "Development Status :: 4 - Beta",
            "Intended Audience :: Healthcare Industry",
            "Programming Language :: Python :: 3",
            "Operating System :: Microsoft :: Windows",
        ],
        license="MIT",
        python_requires=">=3.11",
)



"""
contains multiple classes for various tasks: \n
-Page class for keeping track of page information and annotations \n
-PDF class for reading and modifying pdf files \n
-Annotation class for acessing annotations easily
"""
from __future__ import annotations # Nessecary for typehinting
import os
import logging as lg
import PyPDF2
from PyPDF2.generic import AnnotationBuilder, NameObject, DictionaryObject, RectangleObject
from PyPDF2._page import PageObject


SEPARATORS: tuple[str] = (",", ";", "|") # expand as needed

lg.basicConfig(
    filename=f"{os.path.dirname(__file__)}/Annotation_Exporter.log",
    encoding="utf-8",
    level=lg.DEBUG,
    filemode="w")


class Page:
    """
    Keeps track of the datasets on each page and the page number.
    """
    def __init__(self, page: PageObject, page_nr: int) -> None:
        """
        Initialise class. page_nr is passed because of the way PyPDF2 works
        where the page number is not in the page object.
        
        :param page: The page object.
        :type page: PageObject
        :param page_nr: Page number.
        :type page: int
        """
        self.page: PageObject = page
        self.page_nr: int = page_nr
        self.datasets: list[tuple] = []
        self.has_annotations: bool = False
        self.annotations: list[Annotation] = self.generate_annotation_list()

    def generate_annotation_list(self) -> list[Annotation]:
        """
        Generates the list of annotations, resolves 
        IndirectObject references and creates Annotation objects.
        
        :return: list of annotations
        :rtype: list[Annotation]
        """
        if "/Annots" not in self.page:
            return []

        self.has_annotations = True
        annotation_dictionary_objects: list[DictionaryObject] = [annot.get_object() for annot in self.page["/Annots"]]

        return [
            Annotation(annot_dict, self)
            for dict_obj in annotation_dictionary_objects
            for annot_dict in Annotation.get_multiple_variables(dict_obj)
            ]

    def add_annotation(self, annotation: Annotation) -> None:
        """
        Adds an annotation to the page.

        :param annotation: The annotation to add.
        :type annotation: Annotation
        """
        self.annotations.append(annotation)

    def get_annotations(self) -> list[Annotation]:
        """
        getter for page annotations

        :return: list of annotations
        :rtype: list[Annotation]
        """
        return self.annotations

    def get_page_nr(self) -> int:
        """
        Simple getter for the page number.
        
        :return: The page number.
        :rtype: int
        """
        return self.page_nr

    def get_datasets(self) -> list[tuple]:
        """
        Simple getter for all datasets.

        :return: The datasets. each dataset is a tuple with dataset name and color
        :rtype: list[tuple]
        """
        return self.datasets

    def add_datasets(self, data: tuple) -> None:
        """
        Adds the datasets to the list.

        :param data: The dataset to add.
        :type data: tuple
        """
        self.datasets.append(data)

class PDF:
    """
    Modifies the pdf file and generates a data structure to work on
    """
    def __init__(self, pdf_reader: PyPDF2.PdfReader) -> None:
        self.pdf_reader: PyPDF2.PdfReader =  pdf_reader
        self.pages: list[Page] = self.init_pages()

    def init_pages(self) -> list[Page]:
        """
        Initialises the pages that will initialize the Annotations

        :return: list of pages
        :rtype: list[Page]
        """
        page_list: list[Page] = []
        for page in self.pdf_reader.pages:
            page_list.append(Page(page, self.pdf_reader.get_page_number(page)))

        return page_list

    def convert_old_standard(self, output_folder: str) -> None:
        """
        Converts the old standard to the new standard,
        for reference check the gitHub repo.

        :param output_folder: path to the output folder
        :type output_folder: str
        """
        writer = PyPDF2.PdfWriter()
        new_pdf_path: str = f"{output_folder}/output.pdf"

        writer.append_pages_from_reader(self.pdf_reader)

        for page in self.pages:

            for annot in page.get_annotations():
                if not annot.dataset or annot.new_datset:
                    continue

                dataset_long_name: str = annot.content.split("=", 1)[1].lstrip()
                new_annot = AnnotationBuilder.free_text(
                    f"{annot.dataset_name} ({dataset_long_name})",
                    rect=annot.rect,
                )
                new_annot[NameObject("/C")] = annot.color
                writer.add_annotation(page.get_page_nr(), new_annot)

        with open(new_pdf_path, "wb") as fp:
            writer.write(fp)

class Annotation:
    """
    This is the class that stores a single annotation with convenient 
    access to frequently used proprties and some useful methods
    """
    def __init__(self, annot_obj: DictionaryObject, page: Page) -> None:
        self.is_valid: bool = True
        self.annot_obj = annot_obj
        self.page: Page = page
        self.dataset: bool = False
        self.new_datset: bool = False
        self.supp: bool = False
        self.separators: tuple[str] = SEPARATORS
        self.dataset_name: str = None
        self.assigned_dataset: str = None
        self.variable_name: str = None

        try:
            self.color: list[float] = annot_obj["/C"]
            self.content: str = annot_obj["/Contents"]
            self.content_without_spaces: str = "".join(self.content.split())
            self.rect: RectangleObject = annot_obj["/Rect"]
            self.subtype: str = annot_obj["/Subtype"]
            if self.subtype != "/FreeText":
                raise KeyError # not a KeyError but requires the same action
        except KeyError:
            lg.info("Unsupported Annotation: %s", annot_obj)
            self.is_valid = False
            return

        self.dataset = self.is_dataset()
        self.supp = self.is_supp()

        self.truncate_exess_text()

    def truncate_exess_text(self) -> None:
        """
        truncates the annotation if it contains invalid text
        """
        if self.dataset or self.supp:
            return

        for separator in SEPARATORS:
            if separator not in self.content:
                continue

            self.content = self.content.split(separator, 1)[0]

    @staticmethod
    def get_multiple_variables(annot_obj: DictionaryObject) -> list[dict]:
        """
        returns all variables from an annotation. this is used to pick up on multiple variables
        being in the same annotation.  This has the side effect of converting the DictionaryObjects 
        to dictionaries which looses some (hopefully irrelevant) data.

        :param annot_obj: the annotation object
        :type annot_obj: DictionaryObject
        :return: list of annotations as a list of dictionaries
        :rtype: list[dict]
        """
        split_set = set()
        try: # try except as this is an unsafe annotation
            content: str = annot_obj["/Contents"]
            color: list[float] = annot_obj["/C"]
            subtype: str = annot_obj["/Subtype"]
            rect: list[float] = annot_obj["/Rect"]
        except KeyError:
            lg.info("Unsupported Annotation: %s", annot_obj)
            return []

        for separator in SEPARATORS:
            for possible_variable in content.split(separator):
                if any(ext in possible_variable for ext in SEPARATORS): # if any separator is in the string don't add it
                    continue
                elif "("  in possible_variable: #brackets are special cases
                    possible_variable = possible_variable.split("(", 1)[0]
                elif ")" in possible_variable:
                    possible_variable = possible_variable.split(")", 1)[1]

                if possible_variable == "":
                    continue
                split_set.add(possible_variable)

        split_content = list(split_set)
        return [{"/Contents": string,
                    "/C": color,
                    "/Subtype": subtype,
                    "/Rect": rect}
                    for string in split_content]

    def is_dataset(self) -> bool:
        """
        Returns wehther the annotation is a dataset or not
        and implicitly adds the dataset to the page. There is
        a static method that checks if a string is a dataset 
        aswell so if you need to call this on just a string you
        can use Annotation.is_dataset_static(string).

        :return: if the annotation is a dataset
        :rtype: bool
        """
        if len(self.content_without_spaces.split("(", 1)[0]) == 2: # new standard
            self.new_datset = True
            self.dataset_name = self.content_without_spaces.split("(", 1)[0]
            self.page.add_datasets((self.dataset_name, self.color))
            return True
        elif len(self.content_without_spaces.split("=", 1)[0]) == 2: # old standard
            self.dataset_name = self.content_without_spaces.split("=", 1)[0]
            self.page.add_datasets((self.dataset_name, self.color))
            return True
        self.variable_name = self.content_without_spaces.split("=", 1)[0]
        return False

    def is_supp(self) -> bool:
        """
        returns wether the annotation is a supplementary variable or not

        :return: if the annotation is a supplementary variable
        :rtype: bool
        """
        if self.content_without_spaces.split("=", 1)[0][:4] == "SUPP":
            self.dataset_name = self.content_without_spaces.split("=", 1)[0][:6]
            return True
        return False

    def sort_into_datasets(self) -> None:
        """
        Sorts self into one of the datasets from the page it is on
        """
        for combination in self.page.get_datasets():
            if combination[1] == self.color: # match color
                lg.debug(
                        """Variable %s was assiged %s because %s is the same as %s""", 
                        self.content, combination, self.color, combination[1])
                self.assigned_dataset = combination[0]
                return

        lg.error("no dataset was matched to variable! %s", self)

    @staticmethod
    def is_dataset_static(string: str) -> bool:
        """
        static method that checks if a string is a dataset
        :param string: the string to check

        :type string: str
        :return: True if the string is a dataset, False otherwise
        :rtype: bool
        """
        if len(string.split("(", 1)[0]) == 2: # new standard
            return True
        elif len(string.split("=", 1)[0]) == 2: # old standard
            return True
        return False

    def __str__(self) -> str:
        return f"Annotation: {self.content} on page {self.page.get_page_nr()}"

    def __repr__(self) -> str:
        return f"Annotation: {self.content} on page {self.page.get_page_nr()}"
